from typing import Any

from nonebot import logger

from muicebot.plugin.func_call import get_function_calls
from muicebot.plugin.mcp import handle_mcp_tool


async def function_call_handler(func: str, arguments: dict[str, str] | None = None) -> Any:
    """
    模型 Function Call 请求处理
    """
    arguments = arguments if arguments and arguments != {"dummy_param": ""} else {}

    if func_caller := get_function_calls().get(func):
        logger.info(f"Function call 请求 {func}, 参数: {arguments}")
        result = await func_caller.run(**arguments)
        logger.success(f"Function call 成功，返回: {result}")
        return result

    if mcp_result := await handle_mcp_tool(func, arguments):
        logger.success(f"MCP 工具执行成功，返回: {mcp_result}")
        return mcp_result

    return "(Unknown Function)"



from typing import Dict, Type

from ._base import BaseLLM

LLM_REGISTRY: Dict[str, Type[BaseLLM]] = {}


def register(name: str):
    """
    注册一个 LLM 实现

    :param name: LLM 实现名
    """

    def decorator(cls: Type[BaseLLM]):
        LLM_REGISTRY[name.lower()] = cls
        return cls

    return decorator


def get_llm_class(name: str) -> Type[BaseLLM]:
    """
    获得一个 LLM 实现类

    :param name: LLM 实现名
    """
    if name.lower() not in LLM_REGISTRY:
        raise ValueError(f"未注册模型：{name}")

    return LLM_REGISTRY[name.lower()]



"""first revision

迁移 ID: 8141d1806585
父迁移:
创建时间: 2025-06-05 09:28:35.691503

"""

from __future__ import annotations

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

revision: str = "8141d1806585"
down_revision: str | Sequence[str] | None = None
branch_labels: str | Sequence[str] | None = ("muicebot",)
depends_on: str | Sequence[str] | None = None


def upgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "muicebot_msg",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("time", sa.String(), nullable=False),
        sa.Column("userid", sa.String(), nullable=False),
        sa.Column("groupid", sa.String(), nullable=True, default="-1"),
        sa.Column("message", sa.Text(), nullable=False),
        sa.Column("respond", sa.Text(), nullable=False),
        sa.Column("history", sa.Integer(), nullable=True, default=1),
        sa.Column("resources", sa.Text(), nullable=True, default="[]"),
        sa.Column("usage", sa.Integer(), nullable=True, server_default=sa.text("-1")),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_muicebot_msg")),
        info={"bind_key": "muicebot"},
    )
    # ### end Alembic commands ###


def downgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("muicebot_msg")
    # ### end Alembic commands ###



"""add User table

迁移 ID: 9dd95c5e9b4d
父迁移: 8141d1806585
创建时间: 2025-06-11 21:32:26.115012

"""

from __future__ import annotations

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

revision: str = "9dd95c5e9b4d"
down_revision: str | Sequence[str] | None = "8141d1806585"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "muicebot_user",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("userid", sa.String(), nullable=False),
        sa.Column("nickname", sa.String(), nullable=True, default="_default"),
        sa.Column("profile", sa.String(), nullable=True, default="_default"),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_muicebot_user")),
        info={"bind_key": "muicebot"},
    )
    with op.batch_alter_table("muicebot_msg", schema=None) as batch_op:
        batch_op.add_column(sa.Column("profile", sa.String(), nullable=True, default="_default"))
    op.execute("UPDATE muicebot_msg SET profile = '_default'")

    # ### end Alembic commands ###


def downgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("muicebot_msg", schema=None) as batch_op:
        batch_op.drop_column("profile")

    op.drop_table("muicebot_user")
    # ### end Alembic commands ###



import importlib

from ._base import BaseLLM
from ._config import ModelConfig
from .registry import get_llm_class


def load_model(config: ModelConfig) -> BaseLLM:
    """
    获得一个 LLM 实例
    """
    module_name = config.loader.lower()
    module_path = f"muicebot.llm.providers.{module_name}"

    # 延迟导入模型模块（只导一次）
    importlib.import_module(module_path)

    # 注册之后，直接取类使用
    LLMClass = get_llm_class(config.loader)

    return LLMClass(config)



from nonebot_plugin_orm import Model
from sqlalchemy import Integer, String, Text
from sqlalchemy.orm import Mapped, mapped_column


class Msg(Model):
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    time: Mapped[str] = mapped_column(String, nullable=False)
    userid: Mapped[str] = mapped_column(String, nullable=False)
    groupid: Mapped[str] = mapped_column(String, nullable=True, default="-1")
    message: Mapped[str] = mapped_column(Text, nullable=False)
    respond: Mapped[str] = mapped_column(Text, nullable=False)
    history: Mapped[int] = mapped_column(Integer, nullable=True, default=1)
    resources: Mapped[str] = mapped_column(Text, nullable=True, default="[]")
    usage: Mapped[int] = mapped_column(Integer, nullable=True, default=-1)
    profile: Mapped[str] = mapped_column(String, nullable=True, default="_default")


class User(Model):
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    userid: Mapped[str] = mapped_column(String, nullable=False)
    nickname: Mapped[str] = mapped_column(String, nullable=True, default="_default")
    profile: Mapped[str] = mapped_column(String, nullable=True, default="_default")



import json

import aiosqlite
from arclet.alconna import Alconna
from nonebot import logger
from nonebot.permission import SUPERUSER
from nonebot_plugin_alconna import CommandMeta, on_alconna
from nonebot_plugin_localstore import get_plugin_data_dir
from nonebot_plugin_orm import async_scoped_session

from muicebot.database import MessageORM
from muicebot.models import Message, Resource
from muicebot.plugin import PluginMetadata

__plugin_meta__ = PluginMetadata(
    name="muicebot_plugin_migrations", description="从旧数据库实现中迁移", usage=".migrate"
)

COMMAND_PREFIXES = [".", "/"]

command_migrate = on_alconna(
    Alconna(COMMAND_PREFIXES, "migrate", meta=CommandMeta("从旧数据库实现中迁移")),
    priority=10,
    block=True,
    permission=SUPERUSER,
)


class Database:
    def __init__(self) -> None:
        self.DB_PATH = get_plugin_data_dir().joinpath("ChatHistory.db").resolve()

    def __connect(self) -> aiosqlite.Connection:
        return aiosqlite.connect(self.DB_PATH)

    async def get_version(self) -> int:
        """
        获取数据库版本号，默认值为0
        """
        result = await self.execute("SELECT version FROM schema_version", fetchone=True)
        return result[0] if result else 0

    async def execute(self, query: str, params=(), fetchone=False, fetchall=False) -> list | None:
        """
        异步执行SQL查询，支持可选参数。

        :param query: 要执行的SQL查询语句
        :param params: 传递给查询的参数
        :param fetchone: 是否获取单个结果
        :param fetchall: 是否获取所有结果
        """
        async with self.__connect() as conn:
            conn.row_factory = aiosqlite.Row
            cursor = await conn.cursor()
            await cursor.execute(query, params)
            if fetchone:
                return await cursor.fetchone()  # type: ignore
            if fetchall:
                rows = await cursor.fetchall()
                return [{k.lower(): v for k, v in zip(row.keys(), row)} for row in rows]
            await conn.commit()

        return None

    def connect(self) -> aiosqlite.Connection:
        return aiosqlite.connect(self.DB_PATH)

    async def _deserialize_rows(self, rows: list) -> list[Message]:
        """
        反序列化数据库返回结果
        """
        result = []

        for row in rows:
            data = dict(row)

            # 反序列化 resources
            resources = json.loads(data.get("resources", "[]"))
            data["resources"] = [Resource(**r) for r in resources] if resources else []

            result.append(Message(**data))

        result.reverse()
        return result

    async def get_all_record(self) -> list[Message]:
        query = "SELECT * FROM MSG"
        rows = await self.execute(query, fetchall=True)

        result = await self._deserialize_rows(rows) if rows else []

        return result


@command_migrate.handle()
async def handle_migrate(session: async_scoped_session):
    logger.info("准备从旧数据库实现中迁移...")

    old_db_path = get_plugin_data_dir().joinpath("ChatHistory.db").resolve()

    if not old_db_path.exists():
        msg = "旧数据库文件不存在，停止迁移"
        logger.error(msg)
        await command_migrate.finish(msg)

    old_db = Database()
    old_db_version = await old_db.get_version()

    if old_db_version < 2:
        msg = "旧数据库版本低于 v2，无法迁移"
        logger.error(msg)
        await command_migrate.finish(msg)

    msg_records = await old_db.get_all_record()

    message_orm = MessageORM()

    for record in msg_records:
        await message_orm.add_item(session, record)

    await session.commit()
    msg = f"已成功迁移 {len(msg_records)} 条记录✨"
    logger.success(msg)
    await command_migrate.finish(msg)



from importlib.util import find_spec
from typing import Any, List, Literal, Optional

from pydantic import BaseModel, field_validator


class ModelConfig(BaseModel):
    loader: str = ""
    """所使用加载器的名称，位于 llm 文件夹下，loader 开头必须大写"""

    template: Optional[str] = None
    """使用的人设模板名称"""
    template_mode: Literal["system", "user"] = "system"
    """模板嵌入模式: `system` 为嵌入到系统提示; `user` 为嵌入到用户提示中"""

    max_tokens: int = 4096
    """最大回复 Tokens """
    temperature: float = 0.75
    """模型的温度系数"""
    top_p: float = 0.95
    """模型的 top_p 系数"""
    top_k: float = 3
    """模型的 top_k 系数"""
    frequency_penalty: Optional[float] = None
    """模型的频率惩罚"""
    presence_penalty: Optional[float] = None
    """模型的存在惩罚"""
    repetition_penalty: Optional[float] = None
    """模型的重复惩罚"""
    stream: bool = False
    """是否使用流式输出"""
    online_search: bool = False
    """是否启用联网搜索（原生实现）"""
    function_call: bool = False
    """是否启用工具调用"""
    content_security: bool = False
    """是否启用内容安全"""

    model_path: str = ""
    """本地模型路径"""
    adapter_path: str = ""
    """基于 model_path 的微调模型或适配器路径"""

    model_name: str = ""
    """所要使用模型的名称"""
    api_key: str = ""
    """在线服务的 API KEY"""
    api_secret: str = ""
    """在线服务的 api secret """
    api_host: str = ""
    """自定义 API 地址"""

    extra_body: Optional[dict] = None
    """OpenAI 的 extra_body"""
    enable_thinking: Optional[bool] = None
    """Dashscope 的 enable_thinking"""
    thinking_budget: Optional[int] = None
    """Dashscope 的 thinking_budget"""

    multimodal: bool = False
    """是否为（或启用）多模态模型"""
    modalities: List[Literal["text", "audio", "image"]] = ["text"]
    """生成模态"""
    audio: Optional[Any] = None
    """多模态音频参数"""

    @field_validator("loader")
    @classmethod
    def check_model_loader(cls, loader: str) -> str:
        if not loader:
            raise ValueError("loader is required")

        loader = loader.lower()

        # Check if the specified loader exists
        module_path = f"muicebot.llm.providers.{loader}"

        # 使用 find_spec 仅检测模块是否存在，不实际导入
        if find_spec(module_path) is None:
            raise ValueError(f"指定的模型加载器 '{loader}' 不存在于 llm 目录中")

        return loader



from dataclasses import dataclass, field
from typing import List, Optional

from ..models import Message, Resource


@dataclass
class ModelRequest:
    """
    模型调用请求
    """

    prompt: str
    history: List[Message] = field(default_factory=list)
    resources: List[Resource] = field(default_factory=list)
    tools: Optional[List[dict]] = field(default_factory=list)
    system: Optional[str] = None


@dataclass
class ModelCompletions:
    """
    模型输出
    """

    text: str = ""
    usage: int = -1
    resources: List[Resource] = field(default_factory=list)
    succeed: Optional[bool] = True


@dataclass
class ModelStreamCompletions:
    """
    模型流式输出
    """

    chunk: str = ""
    usage: int = -1
    resources: Optional[List[Resource]] = field(default_factory=list)
    succeed: Optional[bool] = True



import json
from datetime import datetime
from typing import List, Optional

from nonebot_plugin_orm import async_scoped_session
from sqlalchemy import desc, func, select, update

from ..models import Message, Resource
from .orm_models import Msg, User


class MessageORM:
    @staticmethod
    def _convert(row: Msg) -> Message:
        """
        反序列化为 Message 实例
        """
        return Message(
            time=row.time,
            userid=row.userid,
            groupid=row.groupid,
            message=row.message,
            respond=row.respond,
            resources=[Resource(**r) for r in json.loads(row.resources or "[]")],
            usage=row.usage,
            profile=row.profile,
        )

    @staticmethod
    async def add_item(session: async_scoped_session, message: Message):
        """
        将消息保存到数据库
        """
        resources = json.dumps([r.to_dict() for r in message.resources], ensure_ascii=False)
        profile = await UserORM.get_user_profile(session, message.userid)
        session.add(
            Msg(
                time=message.time,
                userid=message.userid,
                groupid=message.groupid,
                message=message.message,
                respond=message.respond,
                resources=resources,
                usage=message.usage,
                profile=profile,
            )
        )

    @staticmethod
    async def get_user_history(session: async_scoped_session, userid: str, limit: int = 0) -> List[Message]:
        """
        获取用户的所有对话历史

        :param userid: 用户名
        :param limit: (可选) 返回的最大长度，当该变量设为0时表示全部返回

        :return: 消息列表
        """
        profile = await UserORM.get_user_profile(session, userid)
        stmt = select(Msg).where(Msg.userid == userid, Msg.history == 1, Msg.profile == profile).order_by(desc(Msg.id))
        if limit:
            stmt = stmt.limit(limit)
        result = await session.execute(stmt)
        rows = result.scalars().all()
        return [MessageORM._convert(msg) for msg in rows][::-1]

    @staticmethod
    async def get_group_history(session: async_scoped_session, groupid: str, limit: int = 0) -> List[Message]:
        """
        获取群组的所有对话历史，返回一个列表

        :param groupid: 群组id
        :param limit: (可选) 返回的最大长度，当该变量设为0时表示全部返回

        :return: 消息列表
        """
        stmt = select(Msg).where(Msg.groupid == groupid, Msg.history == 1).order_by(desc(Msg.id))
        if limit:
            stmt = stmt.limit(limit)
        result = await session.execute(stmt)
        rows = result.scalars().all()
        return [MessageORM._convert(msg) for msg in rows][::-1]

    @staticmethod
    async def mark_history_as_unavailable(
        session: async_scoped_session,
        userid: str,
        limit: Optional[int] = None,
    ):
        """
        将用户消息上下文标记为不可用 (适用于 reset 命令)

        :param userid: 用户id
        :param profile: 消息所属存档
        :param limit: (可选)最大操作数
        """
        profile = await UserORM.get_user_profile(session, userid)
        if limit:
            subq = (
                select(Msg.id)
                .where(Msg.userid == userid, Msg.history == 1, Msg.profile == profile)
                .order_by(desc(Msg.id))
                .limit(limit)
            )
            sub_ids = (await session.execute(subq)).scalars().all()
            if sub_ids:
                await session.execute(update(Msg).where(Msg.id.in_(sub_ids)).values(history=0))
        else:
            await session.execute(update(Msg).where(Msg.userid == userid, Msg.profile == profile).values(history=0))

    @staticmethod
    async def get_model_usage(session: async_scoped_session) -> tuple[int, int]:
        """
        获取模型用量数据（今日用量，总用量）

        :return: today_usage, total_usage
        """
        total = await session.execute(select(func.sum(Msg.usage)).where(Msg.usage != -1))
        today = await session.execute(
            select(func.sum(Msg.usage)).where(Msg.usage != -1, Msg.time.like(f"{datetime.now().strftime('%Y.%m.%d')}%"))
        )
        return (today.scalar() or 0), (total.scalar() or 0)

    @staticmethod
    async def get_conv_count(session: async_scoped_session) -> tuple[int, int]:
        """
        获取对话次数（今日次数，总次数）

        :return: today_count, total_count
        """
        total = await session.execute(select(func.count()).where(Msg.usage != -1))
        today = await session.execute(
            select(func.count()).where(Msg.usage != -1, Msg.time.like(f"{datetime.now().strftime('%Y.%m.%d')}%"))
        )
        return (today.scalar() or 0), (total.scalar() or 0)


class UserORM:
    @staticmethod
    async def create_user(session: async_scoped_session, userid: str) -> User:
        user = User(userid=userid)
        session.add(user)
        await session.commit()
        return user

    @staticmethod
    async def get_user(session: async_scoped_session, userid: str) -> User:
        user = await session.execute(select(User).where(User.userid == userid).limit(1))
        return user.scalar_one_or_none() or await UserORM.create_user(session, userid)

    @staticmethod
    async def set_nickname(session: async_scoped_session, userid: str, nickname: str):
        """
        设置用户昵称

        :param userid: 用户id
        :param nickname: 用户昵称
        """
        await session.execute(update(User).where(User.userid == userid).values(nickname=nickname))

    @staticmethod
    async def set_profile(session: async_scoped_session, userid: str, profile: str = "_default"):
        """
        设置消息存档

        :param userid: 用户id
        :param nickname: 消息存档名
        """
        await session.execute(update(User).where(User.userid == userid).values(profile=profile))

    @staticmethod
    async def get_user_profile(session: async_scoped_session, userid: str) -> str:
        result = await session.execute(select(User.profile).where(User.userid == userid).limit(1))
        profile = result.scalar_one_or_none()
        if profile is not None:
            return profile
        await UserORM.create_user(session, userid)
        return "_default"



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import AsyncGenerator, Literal, Union, overload

from ._config import ModelConfig
from ._schema import ModelCompletions, ModelRequest, ModelStreamCompletions


class BaseLLM(metaclass=ABCMeta):
    """
    模型基类，所有模型加载器都必须继承于该类

    推荐使用该基类中定义的方法构建模型加载器类，但无论如何都必须实现 `ask` 方法
    """

    def __init__(self, model_config: ModelConfig) -> None:
        """
        统一在此处声明变量
        """
        self.config = model_config
        """模型配置"""
        self.is_running = False
        """模型状态"""
        self._total_tokens = -1
        """本次总请求（包括工具调用）使用的总token数。当此值设为-1时，表明此模型加载器不支持该功能"""

    def _require(self, *require_fields: str):
        """
        通用校验方法：检查指定的配置项是否存在，不存在则抛出错误

        :param require_fields: 需要检查的字段名称（字符串）
        """
        missing_fields = [field for field in require_fields if not getattr(self.config, field, None)]
        if missing_fields:
            raise ValueError(f"对于 {self.config.loader} 以下配置是必需的: {', '.join(missing_fields)}")

    def _build_messages(self, request: "ModelRequest") -> list:
        """
        构建对话上下文历史的函数
        """
        raise NotImplementedError

    def load(self) -> bool:
        """
        加载模型（通常是耗时操作，在线模型如无需校验可直接返回 true）

        :return: 是否加载成功
        """
        self.is_running = True
        return True

    async def _ask_sync(self, messages: list) -> "ModelCompletions":
        """
        同步模型调用
        """
        raise NotImplementedError

    def _ask_stream(self, messages: list) -> AsyncGenerator["ModelStreamCompletions", None]:
        """
        流式输出
        """
        raise NotImplementedError

    @overload
    async def ask(self, request: "ModelRequest", *, stream: Literal[False] = False) -> "ModelCompletions": ...

    @overload
    async def ask(
        self, request: "ModelRequest", *, stream: Literal[True] = True
    ) -> AsyncGenerator["ModelStreamCompletions", None]: ...

    @abstractmethod
    async def ask(
        self, request: "ModelRequest", *, stream: bool = False
    ) -> Union["ModelCompletions", AsyncGenerator["ModelStreamCompletions", None]]:
        """
        模型交互询问

        :param request: 模型调用请求体
        :param stream: 是否开启流式对话

        :return: 模型输出体
        """
        pass



from ._base import BaseLLM
from ._config import ModelConfig
from ._dependencies import MODEL_DEPENDENCY_MAP, get_missing_dependencies
from ._schema import ModelCompletions, ModelRequest, ModelStreamCompletions
from .loader import load_model
from .registry import get_llm_class, register

__all__ = [
    "BaseLLM",
    "ModelConfig",
    "ModelRequest",
    "ModelCompletions",
    "ModelStreamCompletions",
    "MODEL_DEPENDENCY_MAP",
    "get_missing_dependencies",
    "register",
    "get_llm_class",
    "load_model",
]
